
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "xstd.h"

#include <iostream.h>
#include <iomanip.h>

#include "Assert.h"
#include "Clock.h"
#include "FileScanner.h"

#include "gadgets.h"

FileScanUserPriority FileScanner::TheMinPriority = fsupMin;


/* FileScanUser */

FileScanUser::~FileScanUser() {
}

// A user may not want to implement read or write
// The default implementation will barf in case the
// corresponding method is actually called
void FileScanUser::noteReadReady(int) {
	Assert(false);
}

void FileScanUser::noteWriteReady(int) {
	Assert(false);
}

void FileScanUser::noteTimeout(int, Time) {
	Assert(false);
}


/* FileScanReserv */

FileScanReserv::FileScanReserv(): theFD(-1), theDir(dirNone), isReady(false) {
}

FileScanReserv::FileScanReserv(int aFD, IODir aDir): theFD(aFD), theDir(aDir), isReady(false) {
}


/* FileScanUserReg */

// note: any activity on the Reg record should reset theStart member
// to upgrade timeout expiration

void FileScanUserReg::reset() {
	theUser = 0;
	theStart = theTimeout = Time();
	theResCount = 0;
	thePriority = fsupDefault;
}

void FileScanUserReg::set(FileScanUser *aUser) {
	Assert(aUser);
	if (theUser) {
		Assert(theUser == aUser);
		Assert(theResCount == 1);
	} else {
		theUser = aUser;
		theTimeout = Time();
	}
	thePriority = fsupDefault;
	theStart = TheClock;
	theResCount++;
}

void FileScanUserReg::clear() {
	Assert(theUser);
	Assert(theResCount > 0);
	if (--theResCount == 0)
		reset();
	else
		theStart = TheClock;
}

void FileScanUserReg::timeout(Time aTout) {
	theTimeout = aTout;
}

void FileScanUserReg::priority(Priority aPriority) {
	thePriority = aPriority;
}

Time FileScanUserReg::waitTime() const {
	return TheClock - theStart;
}

bool FileScanUserReg::timedout() const {
	return thePriority >= FileScanner::TheMinPriority &&
		theTimeout >= 0 && theTimeout <= waitTime();
}

bool FileScanUserReg::needsCheck() const {
	return theUser && thePriority >= FileScanner::TheMinPriority;
}

void FileScanUserReg::noteCheck(bool didIO) {
	if (theUser && didIO) {
		theStart = TheClock;
		thePriority = fsupDefault;
	}
}


/* FileScanner */

FileScanner::FileScanner(): theMaxFD(-1), theResCount(0), theTicker(0) {
}

FileScanner::~FileScanner() {
}

void FileScanner::configure(int fdLimit) {
	theRegs.stretch(fdLimit);
	theRegs.count(fdLimit);
	for (int fd = 0; fd < fdLimit; ++fd)
		theRegs[fd].reset();
}

void FileScanner::ticker(FileScanTicker *aTicker) {
	Assert(!theTicker);
	theTicker = aTicker;
}

void FileScanner::checkTimeouts() {
	for (int fd = 0; fd <= theMaxFD; ++fd) {
		FileScanUserReg &reg = theRegs[fd];
		if (reg.timedout()) {
			if (User *u = reg.user())
				u->noteTimeout(fd, reg.waitTime());
		}
	}
}

int FileScanner::scan(Priority minP, Time *timeout) {

	if (timeout && *timeout < 0) // no time to sweep
		return 0;

	TheMinPriority = minP;

	// silent fix:
	// unfortunately, some selects(2) have magic limits on timeouts
	if (timeout && timeout->tv_sec > 100000000)
		timeout->tv_sec = 100000000;

	//cerr << here << "reserv: " << theResCount << endl;
	const int hotCount = sweep(timeout);
	//cerr << here << "hot files: " << hotCount << endl;

	if (theTicker)
		theTicker->begCycle(hotCount);

	for (int idx = 0; idx < hotCount; ++idx) {
		int fd;
		bool active = false;
		//cerr << here << "idx: " << idx << " ready: " << readyUser(idx, dirRead, fd) << " & " << readyUser(idx, dirWrite, fd) << endl;
		if (User *u = readyUser(idx, dirRead, fd)) {
			theRegs[fd].noteCheck(true);
			u->noteReadReady(fd);
			active = true;
		}
		if (User *u = readyUser(idx, dirWrite, fd)) {
			theRegs[fd].noteCheck(true);
			u->noteWriteReady(fd);
			active = true;
		}

		if (active && theTicker && !theTicker->tick())
			break;
	}

	checkTimeouts();

	if (theTicker)
		theTicker->endCycle();

	return hotCount;
}

FileScanReserv FileScanner::setFD(int fd, IODir dir, FileScanUser *u) {
	theResCount++;
	theRegs[fd].set(u);
	return FileScanReserv(fd, dir);
}

void FileScanner::setTimeout(int fd, Time tout) {
	Assert(fd >= 0 && fd <= theMaxFD);
	theRegs[fd].timeout(tout);
}

void FileScanner::clearTimeout(int fd) {
	theRegs[fd].timeout(Time());
}

void FileScanner::setPriority(int fd, Priority p) {
	theRegs[fd].priority(p);
}

FileScanner::Priority FileScanner::getPriority(int fd) const {
	return theRegs[fd].priority();
}

void FileScanner::clearFD(int fd, IODir) {
	theRegs[fd].clear();
	theResCount--;
}
